/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.osgicdi.impl;

import org.glassfish.osgijavaeebase.Extender;
import org.jboss.weld.bootstrap.WeldBootstrap;
import org.jboss.weld.bootstrap.api.Environments;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.service.packageadmin.PackageAdmin;
import org.osgi.util.tracker.BundleTracker;
import org.osgi.util.tracker.BundleTrackerCustomizer;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

/**
 * An extender that registers a BundleTracker to trigger Weld SPI directly
 * Pl.see http://java.net/jira/browse/GLASSFISH-19215
 *
 * @author Tang Yong(tangyong@cn.fujitsu.com)
 */
public class OSGiCDIExtender implements Extender{
	
    private static final Logger logger =
            Logger.getLogger(OSGiCDIExtender.class.getPackage().getName());
    private BundleContext context;
    private BundleTracker tracker;
    private HashMap<Long, WeldContainerInfo> managedBootstraps = new HashMap<Long, WeldContainerInfo>();

    public OSGiCDIExtender(BundleContext context) {
        this.context = context;
    }

    public synchronized void start() {
    	//executorService = Executors.newSingleThreadExecutor();
    	tracker = new BundleTracker(context, Bundle.ACTIVE | Bundle.STARTING | Bundle.STOPPING, new BeanBundleTrackerCustomizer());
        tracker.open();
    }

    public synchronized void stop() {   	
    	
    	if (tracker != null) tracker.close();
        tracker = null;
        
        Iterator<Entry<Long, WeldContainerInfo>> iterator = managedBootstraps.entrySet().iterator();
        
        while(iterator.hasNext()){
        	Entry<Long, WeldContainerInfo> entry = iterator.next();
        	shutdownManagedBootstrap(entry.getValue());
        	iterator.remove();
        }
        
        managedBootstraps = null;       
    }

    //[Siva's comment]
    //We can do some refactoring to ensure that all the activities 
    //(resetting classloader, shutting down bootstrap, unregistering service registrations etc)
    //that happen when a bundle is stopped or the extender is stopped in one place. Today it is 
    //duplicated in Extender.stop and BundleTrackerCustomizer.modifedBundle. 
	private void shutdownManagedBootstrap(WeldContainerInfo containerInfo) {
		//Shutting Down Weld Container which belongs to current bundle
		ClassLoader oldTCC = Thread.currentThread().getContextClassLoader();
		try{
			ClassLoader newTCC = containerInfo.getClassLoader();
			Thread.currentThread().setContextClassLoader(newTCC);
			containerInfo.getBootstrap().shutdown();
		}catch(RuntimeException e){
			e.printStackTrace();
		}finally{
			Thread.currentThread().setContextClassLoader(oldTCC);
		}	
	}
    
    private static class TCCLClassLoader extends ClassLoader {
        private final Bundle bundle;

        private final ClassLoader infra;

        public TCCLClassLoader(Bundle bundle, ClassLoader infraClassLoader) {
            this.bundle = bundle;
            this.infra = infraClassLoader;
        }

        @Override
        public Class<?> loadClass(String name) throws ClassNotFoundException {
            Class<?> loadedClass = null;
            try {
                loadedClass = bundle.loadClass(name);
            } catch (ClassNotFoundException cnfe) {
                loadedClass = infra.loadClass(name);
            }
            return loadedClass;
        }

        @SuppressWarnings("unchecked")
		@Override
        public Enumeration<URL> getResources(String s) throws IOException {
            Set<URL> urls = new HashSet<URL>();
            Enumeration<?> enumBundle = bundle.getResources(s);
            Enumeration<?> enumInfra = infra.getResources(s);
            if (enumBundle != null) {
                urls.addAll((Collection<? extends URL>) Collections.list(enumBundle));
            }
            if (enumInfra != null) {
                urls.addAll((Collection<? extends URL>) Collections.list(enumInfra));
            }
            return Collections.enumeration(urls);
        }
    }
    
    private class WeldContainerInfo{
    	private WeldBootstrap bootstrap;
    	private ClassLoader cl;
    	
    	public WeldContainerInfo(WeldBootstrap bootstrap, ClassLoader cl){
    		this.bootstrap = bootstrap;
    		this.cl = cl;
    	}
    	
    	public WeldBootstrap getBootstrap(){
    		return bootstrap;
    	}
    	
    	public ClassLoader getClassLoader(){
    		return cl;
    	}
    }
    
    private class BeanBundleTrackerCustomizer implements BundleTrackerCustomizer {

		@Override
		public Object addingBundle(final Bundle bundle, BundleEvent event) {			
			if ( isPlainOSGiBundle(bundle) ){
				if ( isGlassfishCDIEnabled(bundle) && isCDIBundle(bundle) ){
					final int state = bundle.getState();
					if ( isReady(event, state) ) {
						startWeldIntegration();
						WeldBootstrap bootstrap = new WeldBootstrap();
    		            OSGiBeanDeploymentArchiveFactory factory = new OSGiBeanDeploymentArchiveFactory();
    		            OSGiBeanDeployment deployment = new OSGiBeanDeployment(bundle, bootstrap, factory);
    		            
    		            ClassLoader oldTCC = Thread.currentThread().getContextClassLoader();   					
    					try{
    						ClassLoader newTCC = new TCCLClassLoader(bundle, this.getClass().getClassLoader());
    						Thread.currentThread().setContextClassLoader(newTCC);
    						
    						//Start the weld container
    						bootstrap.startContainer(Environments.SE, deployment);
    						//Initialize the weld container
    						bootstrap.startInitialization();
    						//Deploying Beans
    						bootstrap.deployBeans();  		
    						bootstrap.validateBeans();
    						bootstrap.endInitialization();	
        		            
    						WeldContainerInfo wci = new WeldContainerInfo(bootstrap, newTCC);
    						Long bundleId = Long.valueOf(bundle.getBundleId()); 
        		            managedBootstraps.put(bundleId, wci);
        		            
    					}catch(RuntimeException e){
    						e.printStackTrace();
    					}finally{
    						Thread.currentThread().setContextClassLoader(oldTCC);
    					}
																				
		                return bundle;
		            }
				}
			}
			
            return null;
		}
		
		private void startWeldIntegration(){
			// Seeing GLASSFISH-19422			
			PackageAdmin pa = (PackageAdmin) context.getService(context.getServiceReference(PackageAdmin.class.getName()));
			Bundle bundle = pa.getBundles(Constants.WEDINTEGRATION_SYMNAME, null)[0];
			if (bundle.getState() != Bundle.ACTIVE){
				try {
					//bundle.start(Bundle.START_TRANSIENT);
					//Temp Fixing way, please seeing http://java.net/jira/browse/GLASSFISH-19326
					//And https://issues.apache.org/jira/browse/FELIX-3713
					//Once FELIX-3713 is fixed, bundle.start(Bundle.START_TRANSIENT) will be used
					bundle.start();
				} catch (BundleException e) {
					e.printStackTrace();
				}
			}
		}

		private boolean isGlassfishCDIEnabled(Bundle bundle){
			Object gfcdiFlag = bundle.getHeaders().get(Constants.GLASSFISH_CDIENABLED);
			if (gfcdiFlag != null && "true".equals(gfcdiFlag)) {
		            return true;
		    }
			
			return false;
		}
		
		private boolean isPlainOSGiBundle(Bundle bundle){
			final Dictionary<?, ?> headers = bundle.getHeaders();
			
			return headers.get(Constants.EXPORT_EJB) == null && 
	        		headers.get(Constants.WEB_CONTEXT_PATH) == null && 
	                  headers.get(org.osgi.framework.Constants.FRAGMENT_HOST) == null;
		}
		
		private boolean isCDIBundle(Bundle bundle) {
	        URL url = bundle.getEntry(Constants.BEANS_XML_PATH);
	        return url != null;
		}

		private boolean isReady(BundleEvent event, int state) {
			return state == Bundle.ACTIVE ||
                    (state == Bundle.STARTING && (event != null && event.getType() == BundleEvent.LAZY_ACTIVATION));
		}

		@Override
		public void modifiedBundle(Bundle bundle, BundleEvent event,
				Object object) {
			if ( BundleEvent.STOPPING == event.getType() ){
				Long bundleId = Long.valueOf(bundle.getBundleId());
				
				if ( managedBootstraps.containsKey(bundleId) ){
					shutdownManagedBootstrap(managedBootstraps.get(bundleId));
					managedBootstraps.remove(bundleId);
				}
			}
		}

		@Override
		public void removedBundle(Bundle bundle, BundleEvent event,
				Object object) {	
		}	
    }
}
